/* DOSMOD - DOS-prevention apache module
   (c) 2009 Pavel Shved
   Licensed under GPLv2 or any later version.
   This is free software.  There is NO warranty;
   not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "httpd.h"
#include "http_config.h"

#include <time.h>

#include <algorithm>
#include <map>
#include <vector>

#ifdef __cplusplus
extern "C" {
#endif

// This amount, of course, should be loaded from config file.
unsigned const MAX_REQUESTS = 10;

// Key type to keep IP addresses in request hash
typedef std::vector<char> IP_Key;

// Total number of requests from one IP address
static std::map<IP_Key, unsigned int> requests_from;
// mutex for access to the map above
static apr_thread_mutex_t* requests_lock = NULL;


static int mod_dosmod_main_handler(request_rec *r)
{
	apr_sockaddr_t const& remote_sock = *r->connection->remote_addr;
	IP_Key remote_ip = IP_Key(
		(char const*)(remote_sock.ipaddr_ptr),
		(char const*)(remote_sock.ipaddr_ptr)+remote_sock.addr_str_len
	);
	// Critical section: access to shared map
	// (since one of the threads may write, we should protect reads as well).
	apr_thread_mutex_lock(requests_lock);
	bool drop = requests_from[remote_ip]++ >= MAX_REQUESTS;
	apr_thread_mutex_unlock(requests_lock);
	/* NOTE that return code checking should be here.  But since I don't know how to _handle_ errors within Apache return code check is useless. */

	if (drop){
		// Drop client.  Say Apache that this module has "fulfilled" the request
		// TODO: Apache doesn't seem to let 666 pass and returns 500 instead.  Didn't try to avoid it.
		return 666;
	}else{
		// number of requests is OK, passing request further
		return DECLINED;
	}
}

static void mod_dosmod_register_hooks(apr_pool_t *p)
{
	apr_thread_mutex_create(&requests_lock, APR_THREAD_MUTEX_UNNESTED, p);
	ap_hook_handler(mod_dosmod_main_handler, NULL, NULL, APR_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA dosmod_module = {
	STANDARD20_MODULE_STUFF,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	mod_dosmod_register_hooks,
};

#ifdef __cplusplus
} //extern "C"
#endif

